from __future__ import annotations
import asyncio
import contextlib
from typing import Callable, Mapping
from urllib.parse import urljoin
import httpx
from .base import validate, BaseClient, Graph, Proxy, Remote, Resource


class AsyncClient(BaseClient, httpx.AsyncClient):  # type: ignore
    def run(self, name: str, *args, **kwargs):
        """Synchronously call method and run coroutine."""
        return asyncio.new_event_loop().run_until_complete(getattr(self, name)(*args, **kwargs))


class AsyncResource(AsyncClient):
    """An `AsyncClient` which returns json content and has syntactic support for requests."""

    client = property(AsyncClient.clone, doc="upcasted `AsyncClient`")
    __getattr__ = AsyncClient.__truediv__
    __getitem__ = AsyncClient.get
    content_type = Resource.content_type
    __call__ = Resource.__call__

    async def request(self, method, path, **kwargs):
        """Send request with path and return processed content."""
        response = (await super().request(method, path, **kwargs)).raise_for_status()
        if self.content_type(response) == 'json':
            return response.json()
        return response.text if response.encoding else response.content

    async def updater(self, path='', **kwargs):
        response = (await super().request('GET', path, **kwargs)).raise_for_status()
        kwargs['headers'] = dict(kwargs.get('headers', {}), **validate(response))
        yield await self.put(path, (yield response.json()), **kwargs)

    @contextlib.asynccontextmanager
    async def updating(self, path: str = '', **kwargs):
        """Context manager to GET and conditionally PUT json data."""
        updater = self.updater(path, **kwargs)
        json = await updater.__anext__()
        yield json
        await updater.asend(json)

    async def update(self, path: str = '', callback: Callable | None = None, **json):
        """PATCH request with json params.

        Args:
            callback: optionally update with GET and validated PUT.
                `callback` is called on the json result with keyword params, i.e.,
                `dict` correctly implements the simple update case.
        """
        if callback is None:
            return await self.patch(path, json)
        updater = self.updater(path)
        return await updater.asend(callback(await updater.__anext__(), **json))

    async def authorize(self, path: str = '', **kwargs) -> dict:
        """Acquire oauth access token and set `Authorization` header."""
        method = 'GET' if {'json', 'data'}.isdisjoint(kwargs) else 'POST'
        result = await self.request(method, path, **kwargs)
        self.headers['authorization'] = f"{result['token_type']} {result['access_token']}"
        self._attrs['headers'] = self.headers
        return result


class AsyncRemote(AsyncClient):
    """An `AsyncClient` which defaults to posts with json bodies, i.e., RPC.

    Args:
        url: base url for requests
        json: default json body for all calls
        **kwargs: same options as `AsyncClient`
    """

    client = AsyncResource.client
    __getattr__ = AsyncResource.__getattr__
    check = staticmethod(Remote.check)

    def __init__(self, url: str, json: Mapping = {}, **kwargs):
        super().__init__(url, **kwargs)
        self.json = dict(json)

    @classmethod
    def clone(cls, other, path=''):
        return AsyncClient.clone.__func__(cls, other, path, json=other.json)

    async def __call__(self, path='', **json):
        """POST request with json body and check result."""
        response = (await self.post(path, json=dict(self.json, **json))).raise_for_status()
        return self.check(response.json())


class AsyncGraph(AsyncRemote):
    """An `AsyncRemote` client which executes GraphQL queries."""

    Error = httpx.HTTPError
    execute = Graph.execute
    check = classmethod(Graph.check.__func__)  # type: ignore


class AsyncProxy(AsyncClient):
    """An extensible embedded proxy client to multiple hosts.

    The default implementation provides load balancing based on active connections.
    It does not provide error handling or retrying.

    Args:
        *urls: base urls for requests
        **kwargs: same options as `AsyncClient`
    """

    Stats = Proxy.Stats
    priority = Proxy.priority
    choice = Proxy.choice

    def __init__(self, *urls: str, **kwargs):
        super().__init__('https://proxies', **kwargs)
        self.urls = {(url.rstrip('/') + '/'): self.Stats() for url in urls}

    @classmethod
    def clone(cls, other, path=''):
        urls = (urljoin(url, path) for url in other.urls)
        return cls(*urls, trailing=other.trailing, **other._attrs)

    async def request(self, method, path, **kwargs):
        """Send request with relative or absolute path and return response."""
        url = self.choice(method)
        with self.urls[url] as stats:
            response = await super().request(method, urljoin(url, path), **kwargs)
        stats.add(failures=int(response.status_code >= 500))
        return response



import asyncio
import json
import operator
import httpx
import pytest
import clients


@pytest.mark.asyncio
async def test_client(url):
    client = clients.AsyncClient(url, params={'q': 0})
    coros = (
        client.head(),
        client.options(),
        client.post('post'),
        client.put('put'),
        client.patch('patch'),
        client.delete('delete'),
    )
    for coro in coros:
        r = await coro
        assert r.status_code == 200 and r.url.query == b'q=0'
    r = await (client / 'ip').get(params={'q': 1})
    assert set(r.json()) == {'origin'} and r.url.query == b'q=1'


@pytest.mark.asyncio
async def test_resource(url):
    params = {'etag': 'W/0', 'last-modified': 'now'}
    resource = clients.AsyncResource(url, params=params)
    assert isinstance(await resource['encoding/utf8'], str)
    assert isinstance(await resource('stream-bytes/1'), (str, bytes))
    assert (await resource.update('patch', key='value'))['json'] == {'key': 'value'}
    with pytest.raises(httpx.HTTPError, match='404'):
        await resource.status('404')
    with pytest.raises(httpx.HTTPError):
        await resource.update('response-headers', callback=dict, key='value')
    with pytest.raises(httpx.HTTPError):
        async with resource.updating('response-headers') as data:
            assert data['etag'] == 'W/0'


@pytest.mark.asyncio
async def test_content(url):
    resource = clients.AsyncResource(url)
    resource.content_type = lambda response: 'json'
    coro = resource.get('robots.txt')
    assert not hasattr(coro, '__aenter__')
    with pytest.raises(ValueError):
        await coro


def test_authorize(url, monkeypatch):
    resource = clients.AsyncResource(url)
    future = asyncio.Future(loop=asyncio.new_event_loop())
    future.set_result({'access_token': 'abc123', 'token_type': 'Bearer', 'expires_in': 0})
    monkeypatch.setattr(clients.AsyncResource, 'request', lambda *args, **kwargs: future)
    for key in ('params', 'data', 'json'):
        assert resource.run('authorize', **{key: {}}) == future.result()
        assert resource.headers['authorization'] == 'Bearer abc123'


@pytest.mark.asyncio
async def test_remote(url):
    remote = clients.AsyncRemote(url, json={'key': 'value'})
    assert (await remote('post'))['json'] == {'key': 'value'}
    clients.AsyncRemote.check = operator.methodcaller('pop', 'json')
    assert await (remote / 'post')(name='value') == {'key': 'value', 'name': 'value'}


@pytest.mark.asyncio
async def test_graph(url):
    graph = clients.AsyncGraph(url).anything
    data = await graph.execute('{ viewer { login }}')
    assert json.loads(data) == {'query': '{ viewer { login }}', 'variables': {}}
    with pytest.raises(httpx.HTTPError, match='reason'):
        clients.AsyncGraph.check({'errors': ['reason']})


@pytest.mark.asyncio
async def test_proxy(httpbin):
    proxy = clients.AsyncProxy(httpbin.url, f'http://localhost:{httpbin.port}')
    urls = {(await proxy.get('status/500')).url for _ in proxy.urls}
    assert len(urls) == len(proxy.urls)


def test_clones():
    client = clients.AsyncClient('http://localhost/', trailing='/')
    assert str(client) == 'AsyncClient(http://localhost/... /)'
    assert str(client / 'path') == 'AsyncClient(http://localhost/path/... /)'

    resource = clients.AsyncResource('http://localhost/').path
    assert str(resource) == 'AsyncResource(http://localhost/path/... )'
    assert type(resource.client) is clients.AsyncClient

    remote = clients.AsyncRemote('http://localhost/').path
    assert str(remote) == 'AsyncRemote(http://localhost/path/... )'
    assert type(remote.client) is clients.AsyncClient

    proxy = clients.AsyncProxy('http://localhost/', 'http://127.0.0.1') / 'path'
    assert str(proxy) == 'AsyncProxy(https://proxies/... )'



import json
import operator
import io
import httpx
import pytest
import clients


def test_cookies(url):
    client = clients.Client(url, auth=('user', 'pass'), headers={'x-test': 'true'})
    r = client.get('headers', headers={'x-test2': 'true'})
    assert {'x-test', 'x-test2'} <= set(r.request.headers)

    r = client.get('cookies')
    assert r.json() == {'cookies': {}}

    client.get('cookies/set', params={'sessioncookie': '123456789'})
    r = client.get('cookies')
    assert r.json() == {'cookies': {'sessioncookie': '123456789'}}


def test_content(url):
    resource = clients.Resource(url)
    assert resource.get('get')['url'] == url + '/get'
    with pytest.raises(httpx.HTTPError):
        resource.get('status/404')
    assert '<html>' in resource.get('html')
    assert isinstance(resource.get('stream-bytes/10'), bytes)


def test_path(url):
    client = clients.Client(url)
    cookies = client / 'cookies'
    assert isinstance(cookies, clients.Client)
    assert str(cookies.get().url) == url + '/cookies'

    assert str(cookies.get('/').url) == url
    assert str(cookies.get(url).url) == url


def test_trailing(url):
    client = clients.Client(url, trailing='/')
    assert client.get('ip').status_code == 404


def test_syntax(url):
    resource = clients.Resource(url, follow_redirects=True)
    assert set(resource['get']) == {'origin', 'headers', 'args', 'url'}
    resource['put'] = {}
    del resource['delete']

    assert '200' in resource.status
    assert '404' not in resource.status
    assert [line['id'] for line in resource / 'stream/3'] == [0, 1, 2]
    assert next(iter(resource / 'html')) == '<!DOCTYPE html>'
    assert resource('cookies/set', name='value') == {'cookies': {'name': 'value'}}


def test_methods(url):
    resource = clients.Resource(url)
    assert sum(map(len, resource.stream(path='stream-bytes/256'))) == 256
    assert resource.update('patch', name='value')['json'] == {'name': 'value'}
    assert resource.create('post', {'name': 'value'}) is None
    file = resource.download(io.BytesIO(), 'image/png')
    assert file.tell()


def test_authorize(url, monkeypatch):
    resource = clients.Resource(url)
    result = {'access_token': 'abc123', 'token_type': 'Bearer', 'expires_in': 0}
    monkeypatch.setattr(clients.Resource, 'request', lambda *args, **kwargs: result)
    for key in ('params', 'data', 'json'):
        assert resource.authorize(**{key: {}}) == result
        assert resource.headers['authorization'] == 'Bearer abc123'


def test_callback(url):
    resource = clients.Resource(url, params={'etag': 'W/0', 'last-modified': 'now'})
    with pytest.raises(httpx.HTTPError, match='405') as exc:
        resource.update('response-headers', callback=dict, name='value')
    headers = exc.value.request.headers
    assert headers['if-match'] == 'W/0' and headers['if-unmodified-since'] == 'now'
    with pytest.raises(httpx.HTTPError, match='405') as exc:
        with resource.updating('response-headers') as data:
            data['name'] = 'value'
    assert b'"name": "value"' in exc.value.request.read()


def test_meta(url):
    client = clients.Client(url)
    response = client.options('get')
    assert not response.is_error and not response.content
    response = client.head('get')
    assert not response.is_error and not response.content
    del response.headers['content-type']
    assert clients.Resource.content_type(response) == ''


def test_remote(url):
    remote = clients.Remote(url, json={'key': 'value'})
    assert remote('post')['json'] == {'key': 'value'}
    clients.Remote.check = operator.methodcaller('pop', 'json')
    assert (remote / 'post')(name='value') == {'key': 'value', 'name': 'value'}


def test_graph(url):
    graph = clients.Graph(url).anything
    data = graph.execute('{ viewer { login }}')
    assert json.loads(data) == {'query': '{ viewer { login }}', 'variables': {}}
    with pytest.raises(httpx.HTTPError, match='reason'):
        clients.Graph.check({'errors': ['reason']})


def test_proxy(httpbin):
    proxy = clients.Proxy(httpbin.url, f'http://localhost:{httpbin.port}')
    urls = {proxy.get('status/500').url for _ in proxy.urls}
    assert len(urls) == len(proxy.urls)


def test_clones():
    client = clients.Client('http://localhost/', trailing='/')
    assert str(client) == 'Client(http://localhost/... /)'
    assert str(client / 'path') == 'Client(http://localhost/path/... /)'

    resource = clients.Resource('http://localhost/').path
    assert str(resource) == 'Resource(http://localhost/path/... )'
    assert type(resource.client) is clients.Client

    remote = clients.Remote('http://localhost/').path
    assert str(remote) == 'Remote(http://localhost/path/... )'
    assert type(remote.client) is clients.Client

    proxy = clients.Proxy('http://localhost/', 'http://127.0.0.1') / 'path'
    assert str(proxy) == 'Proxy(https://proxies/... )'


def test_singleton():
    @clients.singleton('http://localhost/')
    class custom_api(clients.Resource):
        pass  # custom methods

    assert isinstance(custom_api, clients.Resource)
    assert custom_api.url == 'http://localhost/'
